#ifndef Matrix_H
#define Matrix_H

class Matrix
{
public:
	long height;
	long width;
	long frames;
	float *m;//data

	Matrix(long a,long b,long c)//initialization
	{
		height=a;
		width=b;
		frames=c;
	}

	void ensure()
	{
		for(long i=0;i<height;i++)
		{
			for(long j=0;j<width;j++)
			{
				m[i*width+j]=m[2*height*width+i*width+j];
				m[(frames-1)*height*width+i*width+j]=m[(frames-3)*height*width+i*width+j];
			}
		}
		for(long j=0;j<width;j++)
		{
			for(long k=0;k<frames;k++)
			{
				m[k*height*width+j]=m[k*height*width+2*width+j];
				m[k*height*width+(height-1)*width+j]=m[k*height*width+(height-3)*width+j];
			}
		}
		for(long i=0;i<height;i++)
		{
			for(long k=0;k<frames;k++)
			{
				m[k*height*width+i*width]=m[k*height*width+i*width+2];
				m[k*height*width+i*width+width-1]=m[k*height*width+i*width+width-3];
			}
		}
	}		

	void create()//allocate memory
	{
		long bufsize=height*width*frames*sizeof(float);
		m=(float *)malloc(bufsize);
	}

	void setSize(long a,long b,long c)
	{
		height=a;
		width=b;
		frames=c;
	}

	void setM(long i,long j,long k,float x)
	{
		m[k*height*width+i*width+j]=x;
	}

	float getM(long i,long j,long k)
	{
		return m[k*height*width+i*width+j];
	}

	///////////////////////////////////////////////
	//following functions return new matrix
	Matrix plus(Matrix M2)
	{
		Matrix M3(height,width,frames);
		M3.create();
		for(long i=0;i<height*width*frames;i++)
		{
			M3.m[i]=m[i]+M2.m[i];
		}
		return M3;
	}

	Matrix minus(Matrix M2)
	{
		Matrix M3(height,width,frames);
		M3.create();
		for(long i=0;i<height*width*frames;i++)
		{
			M3.m[i]=m[i]-M2.m[i];
		}
		return M3;
	}

	Matrix multiply(Matrix M2)
	{
		Matrix M3(height,width,frames);
		M3.create();
		for(long i=0;i<height*width*frames;i++)
		{
			M3.m[i]=m[i]*M2.m[i];
		}
		return M3;
	}

	Matrix multiply(float a)
	{
		Matrix M3(height,width,frames);
		M3.create();
		for(long i=0;i<height*width*frames;i++)
		{
			M3.m[i]=m[i]*a;
		}
		return M3;
	}

	///////////////////////////////////////////////
	//following functions do not return new matrix
	void plus2self(Matrix M2)
	{
		for(long i=0;i<height*width*frames;i++)
		{
			m[i]=m[i]+M2.m[i];
		}
	}

	void minus2self(Matrix M2)
	{
		for(long i=0;i<height*width*frames;i++)
		{
			m[i]=m[i]-M2.m[i];
		}
	}

	void multiply2self(Matrix M2)
	{
		for(long i=0;i<height*width*frames;i++)
		{
			m[i]=m[i]*M2.m[i];
		}
	}

	void multiply2self(float a)
	{
		for(long i=0;i<height*width*frames;i++)
		{
			m[i]=m[i]*a;
		}
	}

	void oneMinus2self()
	{
		for(long i=0;i<height*width*frames;i++)
		{
			m[i]=1-m[i];
		}
	}

	void square2self()
	{
		for(long i=0;i<height*width*frames;i++)
		{
			m[i]=m[i]*m[i];
		}
	}

	///////////////////////////////////////////////
	//following functions compute differential
	Matrix dx()
	{
		Matrix M(height,width,frames);
		M.create();
		for(long i=0;i<height;i++)
		{
			for(long j=1;j<width-1;j++)
			{
				for(long k=0;k<frames;k++)
				{
					float a=m[k*height*width+i*width+j+1]-m[k*height*width+i*width+j-1];
					M.m[k*height*width+i*width+j]=a/2;
				}
			}
		}
		//boundary
		for(long i=0;i<height;i++)
		{
			for(long k=0;k<frames;k++)
			{
				M.m[k*height*width+i*width]=m[k*height*width+i*width+1]-m[k*height*width+i*width];
				M.m[k*height*width+(i+1)*width-1]=m[k*height*width+(i+1)*width-1]-m[k*height*width+(i+1)*width-2];
			}
		}
		return M;
	}

	Matrix dy()
	{
		Matrix M(height,width,frames);
		M.create();
		for(long i=1;i<height-1;i++)
		{
			for(long j=0;j<width;j++)
			{
				for(long k=0;k<frames;k++)
				{
					float a=m[k*height*width+(i+1)*width+j]-m[k*height*width+(i-1)*width+j];
					M.m[k*height*width+i*width+j]=a/2;
				}
			}
		}
		//boundary
		for(long j=0;j<width;j++)
		{
			for(long k=0;k<frames;k++)
			{
				M.m[k*height*width+j]=m[k*height*width+width+j]-m[k*height*width+j];
				M.m[k*height*width+(height-1)*width+j]=m[k*height*width+(height-1)*width+j]-m[k*height*width+(height-2)*width+j];
			}
		}
		return M;
	}

	Matrix dz()
	{
		Matrix M(height,width,frames);
		M.create();
		for(long i=0;i<height;i++)
		{
			for(long j=0;j<width;j++)
			{
				for(long k=1;k<frames-1;k++)
				{
					float a=m[(k+1)*height*width+i*width+j]-m[(k-1)*height*width+i*width+j];
					M.m[k*height*width+i*width+j]=a/2;
				}
			}
		}
		//boundary
		for(long j=0;j<width;j++)
		{
			for(long i=0;i<height;i++)
			{
				M.m[i*width+j]=m[height*width+i*width+j]-m[i*width+j];
				M.m[(frames-1)*height*width+i*width+j]=m[(frames-1)*height*width+i*width+j]-m[(frames-2)*height*width+i*width+j];
			}
		}
		return M;
	}

	Matrix laplacian()
	{
		Matrix M(height,width,frames);
		M.create();
		for(long i=1;i<height-1;i++)
		{
			for(long j=1;j<width-1;j++)
			{
				for(long k=1;k<frames-1;k++)
				{
					M.m[k*height*width+i*width+j]=(m[(k+1)*height*width+i*width+j]+m[(k-1)*height*width+i*width+j]+m[k*height*width+(i+1)*width+j]+m[k*height*width+(i-1)*width+j]+m[k*height*width+i*width+j+1]+m[k*height*width+i*width+j-1]-6*m[k*height*width+i*width+j])/6;
				}
			}
		}
		//boundary
		for(long i=0;i<height;i++)
		{
			for(long k=0;k<frames;k++)
			{
				M.m[k*height*width+i*width]=M.m[k*height*width+i*width+1];
				M.m[k*height*width+(i+1)*width-1]=M.m[k*height*width+(i+1)*width-2];
			}
		}
		for(long j=0;j<width;j++)
		{
			for(long k=0;k<frames;k++)
			{
				M.m[k*height*width+j]=M.m[k*height*width+width+j];
				M.m[k*height*width+(height-1)*width+j]=M.m[k*height*width+(height-2)*width+j];
			}
		}
		for(long j=0;j<width;j++)
		{
			for(long i=0;i<height;i++)
			{
				M.m[i*width+j]=M.m[height*width+i*width+j];
				M.m[(frames-1)*height*width+i*width+j]=M.m[(frames-2)*height*width+i*width+j];
			}
		}
		return M;
	}

	void gradient(Matrix fx,Matrix fy,Matrix fz)
	{
		for(long i=0;i<height*width*frames;i++)
		{
			m[i]=fx.m[i]*fx.m[i]+fy.m[i]*fy.m[i]+fz.m[i]*fz.m[i];
		}
	}

	///////////////////////////////////////////////
	float max()
	{
		float a=m[0];
		for(long i=0;i<height*width*frames;i++)
		{
			if(m[i]>a) a=m[i];
		}
		return a;
	}

	float min()
	{
		float a=m[0];
		for(long i=0;i<height*width*frames;i++)
		{
			if(m[i]<a) a=m[i];
		}
		return a;
	}

	void normalize()
	{
		float maxV,minV,range;
		maxV=max();
		minV=min();
		range=maxV-minV;
		for(long i=0;i<height*width*frames;i++) 
		{
			m[i]=(m[i]-minV)/range;
		}
	}

	void print()
	{
		for(long i=0;i<height;i++)
		{
			for(long j=0;j<width;j++)
			{
				for(long k=0;k<frames;k++)
				{
					printf(" %6f ",m[k*height*width+i*width+j]);
				}
			}
			printf("\n");
		}
	}
};

#endif