 
#include"stdio.h"
//dx
__global__ void dx(float *f,int height,int width,int frames,float *fx,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 int j=n%width;
	 if(n<height*width*frames)
	 {
		 if(j!=0 && j!=(width-1))
		 {
			 fx[n]=(f[n+1]-f[n-1])/2;
		 }
		 else if(j==0)
		 {
			 fx[n]=(f[n+1]-f[n]);
		 }
		 else
		 {
			 fx[n]=(f[n]-f[n-1]);
		 }
	 }
}

//dy
__global__ void dy(float *f,int height,int width,int frames,float *fy,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 int i=(n%(height*width))/width;
	 if(n<height*width*frames)
	 {
		 if(i!=0 && i!=(height-1))
		 {
			 fy[n]=(f[n+width]-f[n-width])/2;
		 }
		 else if(i==0)
		 {
			 fy[n]=(f[n+width]-f[n]);
		 }
		 else
		 {
			 fy[n]=(f[n]-f[n-width]);
		 }
	 }
}

//dz
__global__ void dz(float *f,int height,int width,int frames,float *fz,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 int k=n/(height*width);
	 if(n<height*width*frames)
	 {
		 if(k!=0 && k!=(frames-1))
		 {
			 fz[n]=(f[n+height*width]-f[n-height*width])/2;
		 }
		 else if(k==0)
		 {
			 fz[n]=(f[n+height*width]-f[n]);
		 }
		 else
		 {
			 fz[n]=(f[n]-f[n-height*width]);
		 }
	 }
}

//gradient
__global__ void gradient(float *fx,float *fy,float *fz,int height,int width,int frames,float *grad,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 if(n<height*width*frames)
	 {
		 grad[n]=fx[n]*fx[n]+fy[n]*fy[n]+fz[n]*fz[n];
	 }
}

//laplacian
__global__ void laplacian(float *f,int height,int width,int frames,float *lap,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 int i=(n%(height*width))/width;
	 int j=n%width;
	 int k=n/(height*width);
	 if(n<height*width*frames)
	 {
		 if(i==0)i++;
		 if(j==0)j++;
		 if(k==0)k++;
		 if(i==height-1)i--;
		 if(j==width-1)j--;
		 if(k==frames-1)k--;
		 lap[n]=(f[(k+1)*height*width+i*width+j]+f[(k-1)*height*width+i*width+j]+f[k*height*width+(i+1)*width+j]+f[k*height*width+(i-1)*width+j]+f[k*height*width+i*width+j+1]+f[k*height*width+i*width+j-1]-6*f[k*height*width+i*width+j])/6;
	 }
}

//minus
__global__ void minus(float *a,float *b,int height,int width,int frames,float *c,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 if(n<height*width*frames)
	 {
		 c[n]=a[n]-b[n];
	 }
}

//plus2self
__global__ void plus2self(float *a,int height,int width,int frames,float *b,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 if(n<height*width*frames)
	 {
		 b[n]=a[n]+b[n];
	 }
}

//minus2self
__global__ void minus2self(float *a,int height,int width,int frames,float *b,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 if(n<height*width*frames)
	 {
		 b[n]=b[n]-a[n];
	 }
}

//multiplyNumber2self
__global__ void multiplyNumber2self(float *a,int height,int width,int frames,float x,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 if(n<height*width*frames)
	 {
		 a[n]=a[n]*x;
	 }
}

//multiplyMatrix2self
__global__ void multiplyMatrix2self(float *a,int height,int width,int frames,float *b,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 if(n<height*width*frames)
	 {
		 b[n]=b[n]*a[n];
	 }
}

//oneMinus2self
__global__ void oneMinus(float *a,int height,int width,int frames,float *b,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 if(n<height*width*frames)
	 {
		 b[n]=1-a[n];
	 }
}

//boundary ensure
__global__ void ensure(float *f,int height,int width,int frames,long gridWidth)    
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 int i=(n%(height*width))/width;
	 int j=n%width;
	 int k=n/(height*width);
	 if(n<height*width*frames)
	 {
		 if(i==0)f[n]=f[n+2*width];
		 else if(i==height-1)f[n]=f[n-2*width];
		 if(j==0)f[n]=f[n+2];
		 else if(j==width-1)f[n]=f[n-2];
		 if(k==0)f[n]=f[n+2*height*width];
		 else if(k==frames-1)f[n]=f[n-2*height*width];
	 }
}

//normalize
__global__ void normalize(float *f,int height,int width,int frames,float minValue,float range,long gridWidth) 
{        
	 long n=blockIdx.x*blockDim.x+threadIdx.x+gridWidth*(blockIdx.y*blockDim.y+threadIdx.y);
	 if(n<height*width*frames)
	 {
		 f[n]=(f[n]-minValue)/range;
	 }
}

extern "C" void CUDAgvf3D(float * h_f, float *h_u,float *h_v,float *h_o,int height,int width,int frames,float mu,int iteration,float minValue,float range)
{           		
	//////////////////////////////////////////////////////
	//memory allocate
	unsigned long size=height*width*frames*sizeof(float);        
	float *f;
	float *u;
	float *v;
	float *o;
	float *fx;
	float *fy;
	float *fz;
	float *grad;//gradient
	float *temp1x;
	float *temp1y;
	float *temp1z;
	float *temp2;//one minus gradient

	cudaMalloc((void**)&f,size);
	cudaMalloc((void**)&u,size);
	cudaMalloc((void**)&v,size);
	cudaMalloc((void**)&o,size);
	cudaMalloc((void**)&fx,size);
	cudaMalloc((void**)&fy,size);
	cudaMalloc((void**)&fz,size);
	cudaMalloc((void**)&grad,size);
	cudaMalloc((void**)&temp1x,size);
	cudaMalloc((void**)&temp1y,size);
	cudaMalloc((void**)&temp1z,size);
	if(cudaMalloc((void**)&temp2,size)==cudaSuccess)
	{
		printf("\nCUDA memory allocation: successful \n");
	}
	else
	{
		printf("\nCUDA memory allocation: failed \n");
		exit(0);
	}

	cudaMemcpy(f,h_f,size,cudaMemcpyHostToDevice);
	//////////////////////////////////////////////////////
	//initialization

	long gridWidth=32768;
	long gridHeight=(width*height*frames+gridWidth-1)/gridWidth;
	dim3 dimBlock(32,16);//number of threads per block is 512
	dim3 dimGrid((gridWidth+32-1)/32,(gridHeight+16-1)/16);

	normalize<<<dimGrid,dimBlock>>>(f,height,width,frames,minValue,range,gridWidth);
	cudaThreadSynchronize();
	

	dx<<<dimGrid,dimBlock>>>(f,height,width,frames,fx,gridWidth);
	dy<<<dimGrid,dimBlock>>>(f,height,width,frames,fy,gridWidth);
	dz<<<dimGrid,dimBlock>>>(f,height,width,frames,fz,gridWidth);
	cudaThreadSynchronize();

	gradient<<<dimGrid,dimBlock>>>(fx,fy,fz,height,width,frames,grad,gridWidth);

	cudaMemcpy(u,fx,size,cudaMemcpyDeviceToDevice);
	cudaMemcpy(v,fy,size,cudaMemcpyDeviceToDevice);
	cudaMemcpy(o,fz,size,cudaMemcpyDeviceToDevice);
	cudaThreadSynchronize();

	multiplyMatrix2self<<<dimGrid,dimBlock>>>(grad,height,width,frames,fx,gridWidth);
	multiplyMatrix2self<<<dimGrid,dimBlock>>>(grad,height,width,frames,fy,gridWidth);
	multiplyMatrix2self<<<dimGrid,dimBlock>>>(grad,height,width,frames,fz,gridWidth);
	cudaThreadSynchronize();

	//from now on fx, fy and fz are no long derivatives, but multiplied by grad
	oneMinus<<<dimGrid,dimBlock>>>(grad,height,width,frames,temp2,gridWidth);

	cudaThreadSynchronize();

	//////////////////////////////////////////////////////
	//iteration
	
	for (int it=0;it<iteration;it++)
	{
		
		ensure<<<dimGrid,dimBlock>>>(u,height,width,frames,gridWidth);
		ensure<<<dimGrid,dimBlock>>>(v,height,width,frames,gridWidth);
		ensure<<<dimGrid,dimBlock>>>(o,height,width,frames,gridWidth);
		cudaThreadSynchronize();
		

		laplacian<<<dimGrid,dimBlock>>>(u,height,width,frames,temp1x,gridWidth);
		laplacian<<<dimGrid,dimBlock>>>(v,height,width,frames,temp1y,gridWidth);
		laplacian<<<dimGrid,dimBlock>>>(o,height,width,frames,temp1z,gridWidth);
		cudaThreadSynchronize();

		multiplyNumber2self<<<dimGrid,dimBlock>>>(temp1x,height,width,frames,mu*4,gridWidth);
		multiplyMatrix2self<<<dimGrid,dimBlock>>>(temp2,height,width,frames,u,gridWidth);
		multiplyNumber2self<<<dimGrid,dimBlock>>>(temp1y,height,width,frames,mu*4,gridWidth);
		multiplyMatrix2self<<<dimGrid,dimBlock>>>(temp2,height,width,frames,v,gridWidth);
		multiplyNumber2self<<<dimGrid,dimBlock>>>(temp1z,height,width,frames,mu*4,gridWidth);
		multiplyMatrix2self<<<dimGrid,dimBlock>>>(temp2,height,width,frames,o,gridWidth);
		cudaThreadSynchronize();

		plus2self<<<dimGrid,dimBlock>>>(temp1x,height,width,frames,u,gridWidth);
		plus2self<<<dimGrid,dimBlock>>>(temp1y,height,width,frames,v,gridWidth);
		plus2self<<<dimGrid,dimBlock>>>(temp1z,height,width,frames,o,gridWidth);
		cudaThreadSynchronize();

		plus2self<<<dimGrid,dimBlock>>>(fx,height,width,frames,u,gridWidth);
		plus2self<<<dimGrid,dimBlock>>>(fy,height,width,frames,v,gridWidth);
		plus2self<<<dimGrid,dimBlock>>>(fz,height,width,frames,o,gridWidth);
		cudaThreadSynchronize();

	}


	//////////////////////////////////////////////////////
	//memory copy and free
	cudaMemcpy(h_u,u,size,cudaMemcpyDeviceToHost);
	cudaMemcpy(h_v,v,size,cudaMemcpyDeviceToHost);
	cudaMemcpy(h_o,o,size,cudaMemcpyDeviceToHost);
	cudaThreadSynchronize();

	cudaFree(f);
	cudaFree(u);
	cudaFree(v);
	cudaFree(o);
	cudaFree(fx);
	cudaFree(fy);
	cudaFree(fz);
	cudaFree(grad);
	cudaFree(temp1x);
	cudaFree(temp1y);
	cudaFree(temp1z);
	cudaFree(temp2);
}